function NOAAWindDemo
%% Demonstrates NOAA Wind Model for obtaining wind data and plotting it.
%  ------------------------------------------------------------------------
%  See also NOAAWindData, MergeNOAAWindData, PlotWindField
%  ------------------------------------------------------------------------
%%
%--------------------------------------------------------------------------
%   Copyright (c) 2010-2012 Princeton Satellite Systems, Inc. 
%   All rights reserved.
%--------------------------------------------------------------------------

%% load city lat/lon data 
%-------------------------
% from the web or a mat-file:
fprintf(1,'Load city lat/lon data...\n')
if( exist('CityLatLonData.mat') )
  latLonData = load('CityLatLonData.mat');
else
  latLonData = CityCodeLatLonParser;
end

%% load current NOAA wind data from the web
%------------------------------------------

% low altitudes
fprintf(1,'Load low-altitude wind data for Chicago region from NOAA website...\n')
dataLo = NOAAWindData( 'chicago', 'lo', latLonData );

% high altitudes
fprintf(1,'Load high-altitude wind data for Chicago region from NOAA website...\n')
dataHi = NOAAWindData( 'chicago', 'hi', latLonData );

% merge low and high data sets
fprintf(1,'Merge low and high altitude wind data into single structure...\n')
data = MergeNOAAWindData( dataLo, dataHi );

%% plot the wind field
%---------------------
fprintf(1,'Plot the 3D wind field.\n')
altRange = [0 20];
fig = PlotWindField( data, altRange );


%--------------------------------------
% PSS internal file version information
%--------------------------------------
% $Date: 2009-11-04 17:50:02 -0500 (Wed, 04 Nov 2009) $
% $Revision: 17946 $
function data = CityCodeLatLonParser( s )

%--------------------------------------------------------------------------
%   Parse a string representation of U.S. city lat/lon data into a MATLAB
%   data structure. The string can be loaded into MATLAB initially by:
%
%   s = urlread('http://www.realestate3d.com/gps/latlong.htm');
%
%   If no string is provided as an argument, this command is called to
%   attempt to load in the string from the website.
%
%   The data structure may be saved to a mat-file for future reference. 
%
%   A portion of the string content is shown below for an example:
%
%     ID     LAT    LON     Location
% 
%    Minnesota
% 
%    [AEL]  43.68   93.37  Albert Lea,MN
%    [AXN]  45.87   95.38  Alexandria,MN
%    [BJI]  47.50   94.93  Bemidji Muni,MN
%    [BRD]  46.40   94.13  Brainerd-Crw,MN
%    [DTL]  46.82   95.88  Detroit Laks,MN
%    [DLH]  46.83   92.18  Duluth,MN
%    [ELO]  47.90   91.82  Ely,MN
%
%--------------------------------------------------------------------------
%   Form:
%   data = CityCodeLatLonParser( s )
%--------------------------------------------------------------------------
%
%   ------
%   Inputs
%   ------
%   s               (:)   String representation of U.S. city lat/lon coordinates
%
%   -------
%   Outputs
%   -------
%   data            (.)   Data structure with fields:
%                         .code  {1,n}    Cell array of city/airport codes
%                         .lat   (1,n)    Vector of latitudes
%                         .lon   (1,n)    Vector of longitudes
%                         .city  {1,n}    Cell array of city names
%                         .state {1,n}    Cell array of state IDs (ie, MN)
%                         .s      (:)     Original string 
%
%--------------------------------------------------------------------------
%	  Web Site Referenced:  http://www.realestate3d.com/gps/latlong.htm
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
%   Copyright 2009 Princeton Satellite Systems, Inc.
%   All rights reserved.
%--------------------------------------------------------------------------

if( nargin < 1 )
   s = urlread('http://www.realestate3d.com/gps/latlong.htm');
   k1 = findstr(s,'ID     LAT    LON     Location');
   k2 = findstr(s,'</PRE>')-1;
   s = s(k1:k2);
end

data.s = s;
data.code = {};
data.lat = [];
data.lon = [];
data.city = {};
data.state = {};

r = s;
k = 0;
while 1
   [t,r] = strtok(r);
   
   if( isempty(t) )
      break;
   end
   
   if( t(1)=='[' & t(5)==']' )
      
      % new line
      k = k+1;
      data.code{k} = t(2:4);
      
      [t,r] = strtok(r);
      data.lat(k) = str2num(t);
      
      [t,r] = strtok(r);
      data.lon(k) = str2num(t);

      kcomma = findstr(r,',');
      data.city{k} = fliplr(deblank(fliplr(r(1:kcomma-1))));
      data.state{k} = r(kcomma+1:kcomma+2);
      
   end
   
end
   
% PSS internal file version information
%--------------------------------------
% $Date: 2012-07-21 08:40:51 -0400 (Sat, 21 Jul 2012) $
% $Revision: 30028 $


function data = MergeNOAAWindData( data1, data2 )

%--------------------------------------------------------------------------
%   Merge two NOAA wind data sets. There are two possibilities:
%     1) Some cities in common, but one is low alt., the other high alt.
%           Here, we append the alt / wind data for each common city.
%     2) No cities in common. Just append like this: data = [data1, data2]
%--------------------------------------------------------------------------
%   Form:
%   data = MergeNOAAWindData( data1, data2 )
%--------------------------------------------------------------------------
%
%   ------
%   Inputs
%   ------
%   data1           (.)   Data set 1 from NOAAWindData
%   data2           (.)   Data set 2 from NOAAWindData
%
%   -------
%   Outputs
%   -------
%   data            (.)   Merged data set
%
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
%   Copyright 2009 Princeton Satellite Systems, Inc.
%   All rights reserved.
%--------------------------------------------------------------------------

n1 = length(data1);
n2 = length(data2);

% create new names. dataA is shorter, dataB is longer
if( n1>n2 )
   dataA = data1;
   n = n1;
   dataB = data2;
else
   dataA = data2;
   n = n2;
   dataB = data1;
end

clear data1 data2

% get data set A codes
nA = length(dataA);
codesA = cell(1,nA);
for i=1:nA
   codesA{i} = dataA(i).code;
end

% get data set B codes
nB = length(dataB);
codesB = cell(1,nB);
for i=1:nB
   codesB{i} = dataB(i).code;
end

% any in common?
allCodes = unique( [codesA, codesB] );
if( length(allCodes) < length(codesA) + length(codesB) )
   common = 1;
else
   common = 0;
end

% if none in common, just append data sets
if( ~common )
   data = [dataA,dataB];
   return;
end

% otherwise... merge data for common cities

% use copy "dataA" for read-only
data = dataA;

for i=1:n
   
   k = strmatch(dataA(i).code,codesB);
   if( ~isempty(k) )
      
      % append
      data(i).h       = [dataA(i).h,      dataB(k).h];
      data(i).hM      = [dataA(i).hM,     dataB(k).hM];
      data(i).hFt     = [dataA(i).hFt,    dataB(k).hFt];
      data(i).heading = [dataA(i).heading,dataB(k).heading];
      data(i).windMag = [dataA(i).windMag,dataB(k).windMag];
      data(i).windN   = [dataA(i).windN,  dataB(k).windN  ];
      data(i).windE   = [dataA(i).windE,  dataB(k).windE  ];

      % sort
      [hs,ks] = sort( data(i).h );
      data(i).hM      = data(i).hM(ks);
      data(i).hFt     = data(i).hFt(ks);
      data(i).heading = data(i).heading(ks);
      data(i).windMag = data(i).windMag(ks);
      data(i).windN   = data(i).windN(ks);
      data(i).windE   = data(i).windE(ks);
      
   end
end
      
      


function data = NOAAWindData( area, lvl, latLonData )

%--------------------------------------------------------------------------
%   Load wind data from the NOAA web site. The web site is:
%
%   http://aviationweather.gov/products/nws/winds/?area=#1&fint=06&lvl=#2
%   http://aviationweather.gov/products/nws/#1?fint=06&lvl=#2
%   
%   where #1 and #2 denote the first and second arguments, 'area' and 'lvl'
%  
%   For more information on the data format, see:
%     http://aviationweather.gov/products/nws/winds/info.php
%
%   NOTE: You must have a working internet connection to use this function.
%
%--------------------------------------------------------------------------
%   Form:
%   data = NOAAWindData( area, lvl )
%--------------------------------------------------------------------------
%
%   ------
%   Inputs
%   ------
%   area            (:)   Geographic area to get wind data from. 
%                         Choose from:
%                          chicago, boston, miami, ft worth, salt lake,
%                          san francisco
%
%   lvl             (:)   Altitude level. Choose from:  lo, hi
%   latLonData      (:)   Data structure array of city lat/lon data
%
%   -------
%   Outputs
%   -------
%   data            (.)   Wind data set with fields:
%                         .code      (:)     3 letter city/airport code
%                         .lat       (1)     Latitude of location
%                         .lon       (1)     Longitude of location
%                         .h        (1,n)    Vector of altitudes
%                         .heading  (1,n)    Vector of wind headings
%                         .windMag  (1,n)    Vector of wind magnitudes
%                         .windN    (1,n)    Vector of wind north components
%                         .windE    (1,n)    Vector of wind east components
%
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
%   Copyright 2009 Princeton Satellite Systems, Inc.
%   All rights reserved.
%--------------------------------------------------------------------------

if( nargin < 3 )
   if( ~exist('CityLatLonData.mat') )
      disp(' ')
      warning('File "CityLatLonData.mat" not found. Latitude / longitude data not entered.');
      disp('Attempting to generate data with CityCodeLatLonParser.')
      disp(' ')
      latLonData = CityCodeLatLonParser;
      return;
   else
      latLonData = load('CityLatLonData');
   end
      
end

% check latLonData
if( isstr(latLonData) )
  if( exist(latLonData,'file') )
    list = load(latLonData);
    if( ~isfield(latLonData,'code') | ~isfield(latLonData,'city') | ...
        ~isfield(latLonData,'lat') | ~isfield(latLonData,'lon') | ...
        ~isfield(latLonData,'lat') | ~isfield(latLonData,'lon') )
      error(sprintf('Loaded data structure from "%s" missing required fields.',latLonData))
    end
  else
    error(sprintf('Attempting to load "%s" but file does not exist.',latLonData))
  end
end


if( nargin < 1 )
   lvl = 'hi';
   area = 'chicago';
end

% lower case
area = lower(area);
lvl  = lower(lvl);

% remove white space
ksp = findstr(area,' ');
area = area(setdiff(1:length(area),ksp));

areas = {'chicago','boston','miami','ftworth','saltlake','sanfrancisco','all'};
lvls  = {'lo','hi'};

if( isempty(strmatch(area,areas,'exact')) )
   disp('WARNING: Unknown area provided. Valid areas to choose from:');
   disp(areas')
   data = [];
   return
end

if( isempty(strmatch(lvl,lvls,'exact')) )
   disp('WARNING: Unknown level. Valid levels to choose from:');
   disp(lvls')
   data = [];
   return
end

% unclear if this can change, but it always seems to be '06'
fint = '06';

url = sprintf('http://aviationweather.gov/products/nws/%s?fint=%s&lvl=%s',...
   area,fint,lvl);
 
s = urlread( url );

switch lvl
   case 'lo'
      k1 = strfind(s,'FT  3000');
   case 'hi'
      k1 = strfind(s,'FT   45000');
end

kend = strfind(s,'</pre>');

s = s(k1:kend-1);

[h,data] = ParseData( s, latLonData );

for i=1:length(data)
  k=strmatch(data(i).code,latLonData.code);
  if( ~isempty(k) )
    data(i).city = latLonData.city{k};
    data(i).state = latLonData.state{k};
  end
end
    

function [h,data] = ParseData( s, latLonData )

% Sample message decoded:
% 
% DATA BASED ON 010000Z
% 
% Forecast data is based on computer forecasts generated the first day of the month at 0000 UTC.
% VALID 010600Z FOR USE 0500-0900Z. TEMPS NEG ABV 24000
% The valid time of the forecast is the first day of the month at 0600 UTC. 
% The forecast winds and temperature are to be used between 0500 and 0900 UTC. 
% Temperatures are negative above 24,000 feet.
% 
% FT 3000 6000 9000 12000 18000 24000 30000 34000 39000
% 
% FT indicates the forecast location, the numbers indicate the forecast levels.
% 
% MKC 2426 2726-09 2826-14 2930-21 2744-32 2751-41 275550 276050 276547
% 
% This example shows data for MKC (Kansas City, MO). 
% The 3,000 foot wind is forecast to be 240 degrees at 26 knots. 
% The 6,000 foot wind is forecast to be 270 degrees at 26 knots and 
% the air temperature is forecast to be -9 degrees Celsius. 
% The 30,000 foot wind is forecast to be 270 degrees at 55 knots with 
% the air temperature forecast to be -50 degrees Celsius.
% 
% Wind direction is coded as a number between 51 and 86 (vice 01 to 36) 
% when the wind speed is 100 knots or greater. 
% To derive the actual wind direction, subtract 50 from the first pair of numbers. 
% To derive wind speed, add 100 to the second pair of numbers. 
% For example, a forecast at 39,000 feet of "731960" shows a wind direction 
% from 230 degrees (73-50=23) with a wind speed of 119 knots (100+19=119). 
% Above 24,000 feet the temperature is assumed to be negative, 
% therefore the third pair of numbers indicate a temperature of minus 60 degrees Celsius.
% 
% If the wind speed is forecast to be 200 knots or greater, 
% the wind group is coded as 199 knots. 
% For example, "7799" is decoded as 270 degrees at 199 knots or greater.
% 
% Wind direction is coded to the nearest 10 degrees. 
% When the forecast speed is less than 5 knots, the coded group is "9900" 
% and read, "LIGHT AND VARIABLE." 

r=s;

% first get altitude vector
h = [];
while 1
   [t,r] = strtok(r);
   if( ~strcmp(t,'FT') )
      htmp = str2num(t);
      if( ~isempty(htmp) )
         h(end+1) = htmp;
      else
         break;
      end
   end
end

% now get wind data
knots2mps = 0.51444444444;
knots2fps = 1.6878099;
ft2m      = 0.3048;
data(1).code = t;
data(1).heading = [];
data(1).windMag = [];
data(1).windN   = [];
data(1).windE   = [];
data(1).h       = h;
data(1).hM      = h*ft2m;
data(1).hFt     = h;
codes{1} = data(1).code;
k = 1;
while 1
   [t,r] = strtok(r);
   if( isempty(t) )
      break;
   end
   
   % heading in digits 1,2
   htmp = str2num(t(1:2));
   
   if( ~isempty(htmp) )
      
      % wind speed in digits 3,4
      wtmp = str2num(t(3:4));
      
      % code 9900 means wind mag less than 5 knots
      if( htmp==99 && wtmp == 00 )
        htmp = 0;
        wmag = 0;
      end
            
      % special case: wind speed greater than 100 knots, heading is 51-86
      if( htmp>=51 )
         htmp = htmp - 50;
         wtmp = wtmp + 100;
      end
      
      % scale heading to give degrees, wind speed to give ft/s
      head = htmp*10;
      wmag = wtmp*knots2mps;
      data(k).heading(end+1) = head;
      data(k).windMag(end+1) = wmag;
      data(k).windN(end+1)   = wmag*cos(pi+head*pi/180);
      data(k).windE(end+1)   = wmag*sin(pi+head*pi/180);
   else
      k = k+1;
      data(k).code = t;
      data(k).heading = [];
      data(k).windMag = [];
      data(k).windN   = [];
      data(k).windE   = [];
      data(k).h       = h;
      data(k).hM      = h*ft2m;
      data(k).hFt     = h;
      codes{k} = data(k).code;
      
   end
end

% some lines do not provide data for the first altitude (3000 ft)
for i=1:length(data), 
   if( length(data(i).heading) < length(data(i).h) )
      z = zeros(1,length(data(i).h)-length(data(i).heading));
      data(i).heading = [z data(i).heading];
      data(i).windMag = [z data(i).windMag];
      data(i).windN   = [z data(i).windN  ];
      data(i).windE   = [z data(i).windE  ];
      if( length(data(i).windMag)~=length(data(i).h) )
         1;
      end
   end
end

for i=1:length(data)
   k = strmatch(data(i).code,latLonData.code);
   if( ~isempty(k) )
      data(i).lat = latLonData.lat(k(1));
      data(i).lon = latLonData.lon(k(1));
   else
      warning(sprintf('Code "%s" not found. Entering 0.0 for lat and lon data.',data(i).code));
      data(i).lat = 0;
      data(i).lon = 0;
   end
end




function fig = PlotWindField( d, altRange, fig )

%--------------------------------------------------------------------------
%   Plot the wind field from a data structure array. Each element of the
%   array has a single lat/lon coordinate, and a vector of wind speeds and
%   headings across an altitude range. 
% 
%   See also: NOAAWindData, MergeMOAAWindData
%--------------------------------------------------------------------------
%   Form:
%   PlotWindField( windData )
%--------------------------------------------------------------------------
%
%   ------
%   Inputs
%   ------
%   d           (.)     Wind data set from NOAAWindData.
%   altRange   (1,2)    Altitude range. Optional. (km)
%   fig         (1)     Figure handle. Optional.
%
%   -------
%   Outputs
%   -------
%   fig         (1)     Figure handle.
%
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
%   Copyright 2009 Princeton Satellite Systems, Inc.
%   All rights reserved.
%--------------------------------------------------------------------------

if( nargin < 3 )
   fig = figure;
else
   figure(fig);
end

% include all altitudes by default
if( nargin < 2 )
   altRange = [0 1e9];
end

% first get lat / lon points, max h
n = length(d);
lat = zeros(1,n);
lon = zeros(1,n);
h = [];
wmagMax = 0;
for i=1:n
   lat(i) = d(i).lat;
   lon(i) = d(i).lon;
   htmp = d(i).h;
   if( length(htmp) > length(h) )
      h = htmp;
   end
   wmagMax = max( wmagMax, max(d(i).windMag) );
end

% negate longitudes, as they are initially measured positive west
% we want to plot with positive east
lon = -lon;

lat = lat*pi/180;
lon = lon*pi/180;

% convert from lat / lon coord to km north / east of reference
Re = 6378.14;
% deg2rad = pi/180;
ft2km   = 0.3048*1e-3;
h = h*ft2km;
% x = lon*deg2rad*Re; %.*cos(lat*deg2rad);
% y = lat*deg2rad*Re;
x = zeros(1,n);
y = zeros(1,n);
for i=2:n,
  [r,az]=LatLonToRA(lat(1),lon(1),lat(i),lon(i),Re);
  x(i)=r*sin(az);
  y(i)=r*cos(az);
end

% if altRange is entered as an index, define surrounding alt range for that
% altitude entry.
if( length(altRange)==1 )
   index = altRange;
   altRange = [h(index)-.1, h(index)+.1];
end

% 2D plot?
k = find( h >= min(altRange) & h <= max(altRange) );
if( length(k)<2 )
   twoDPlot = 1;
else
   twoDPlot = 0;
end

% define scale so that max wind velocity is about 1/10 of horizontal
% distance
scale = (max(x)-min(x))/(2*wmagMax);

scale = scale*1.5;

% load us border map and scale to lat/long coordinates
% g=load('usborder.mat','x','y');
% g.x = g.x*56/max(g.x) - 124;
% g.y = g.y*24/max(g.y) + 24.5;
% 
% plot(g.x,g.y,'r')
% hold on

east = 1;
north = 1;

% create plot
for i=1:n
   if( lat(i)~=0 )
      z = d(i).h*ft2km;
      k = find( z >= min(altRange) & z <= max(altRange) );
      z = z(k);
      wN = d(i).windN(k)*north;
      wE = d(i).windE(k)*east;
      wU = 0*wE;
      
      xs = x(i)*ones(1,length(z));
      ys = y(i)*ones(1,length(z));

      if( twoDPlot )
         %plot(x(i),y(i),'b.')
         quiver(xs,ys,wE,wN,scale);
      else
         %plot3(x(i),y(i),z,'b.')
         quiver3(xs,ys,z,wE,wN,wU,scale);
      end
      
      hold on
   end
end
grid on
if( twoDPlot )
   zoom on
else
   rotate3d on
end

xlabel('East (km)')
ylabel('North (km)')
zlabel('Altitude (km)')


function [range,az]=LatLonToRA(lat0, lon0, latF, lonF, Rp)

% spherical trignometry to set up solution
A = lonF-lon0;
b = pi/2-lat0;
c = pi/2-latF;

% range
cosa = cos(b)*cos(c)+sin(b)*sin(c).*cos(A);
angleDist = acos( cosa );
range = Rp*angleDist;

% azimuth angle
az = atan2( cos(lat0)*sin(A).*cos(latF), sin(latF)-cosa*sin(lat0) );






